#define UTS_VERSION "#11 Fri Apr 24 23:03:10 JST 1998"
#define LINUX_COMPILE_TIME "23:03:10"
#define LINUX_COMPILE_BY "somebody"
#define LINUX_COMPILE_HOST "unknown"
#define LINUX_COMPILE_DOMAIN "somewhere.org"
#define LINUX_COMPILER "gcc version 2.7.2.3"
