/*
 * Copyright (C) 2022 Free Software Foundation, Inc.
 *
 * This file is part of GNU Mach.
 *
 * GNU Mach is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef _UTIL_BYTEORDER_H_
#define _UTIL_BYTEORDER_H_

#include <stdint.h>

uint16_t ntohs(uint16_t netshort);
uint32_t ntohl(uint32_t netlong);

uint16_t htons(uint16_t hostshort);
uint32_t htonl(uint32_t hostlong);

#endif	/* _UTIL_BYTEORDER_H_ */
